package com.loafle.commons.server;

import java.net.SocketAddress;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.ServerChannel;
import io.netty.channel.nio.NioEventLoopGroup;

/**
 * Server
 */
public abstract class Server {
  private static final Logger logger = LoggerFactory.getLogger(Server.class);

  @Value("${server.netty.thread.count.boss}")
  protected int threadCountBoss;
  @Value("${server.netty.thread.count.worker}")
  protected int threadCountWorker;

  @Autowired
  protected List<Class<? extends ServerChannel>> channelClasses;

  @Autowired
  protected List<ChannelOptionItem<?>> channelOptions;
  @Autowired
  protected List<ChannelHandler> channelHandlers;
  @Autowired
  protected ChannelInitializer<?> channelInitializer;
  @Autowired
  protected SocketAddress address;

  public void start() throws Exception {
    EventLoopGroup bossGroup = new NioEventLoopGroup(this.threadCountBoss);
    EventLoopGroup workerGroup = new NioEventLoopGroup(this.threadCountWorker);

    try {
      ServerBootstrap serverBootstrap = new ServerBootstrap();
      serverBootstrap.group(bossGroup, workerGroup);

      if (null != this.channelClasses) {
        for (Class<? extends ServerChannel> channelClass : this.channelClasses) {
          serverBootstrap.channel(channelClass);
        }
      }
      if (null != this.channelOptions) {
        for (ChannelOptionItem<?> optionItem : this.channelOptions) {
          optionItem.setOption(serverBootstrap);
        }
      }
      if (null != this.channelHandlers) {
        for (ChannelHandler channelHandler : this.channelHandlers) {
          serverBootstrap.handler(channelHandler);
        }
      }
      serverBootstrap.childHandler(this.channelInitializer);

      this.init();

      ChannelFuture cf = this.bind(serverBootstrap);
      cf.sync();
      this.onStart();

      cf.channel().closeFuture().sync();

      this.onStop();
    } catch (Exception e) {
      logger.error("Server", e);
    } finally {
      bossGroup.shutdownGracefully();
      workerGroup.shutdownGracefully();
      this.destroy();
    }
  }

  protected ChannelFuture bind(ServerBootstrap serverBootstrap) throws Exception {
    return serverBootstrap.bind(address);
  }

  protected abstract void init() throws Exception;
  protected abstract void onStart() throws Exception;
  protected abstract void onStop() throws Exception;
  protected abstract void destroy() throws Exception;

  public static class ChannelOptionItem<T> {
    private final ChannelOption<T> option;
    private final T value;

    public ChannelOptionItem(ChannelOption<T> option, T value) {
      this.option = option;
      this.value = value;
    }
    void setOption(ServerBootstrap serverBootstrap) {
      serverBootstrap.option(option, value);
    }
  }
}